/*
 * Copyright (c) 2023 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */

#include "streamctrl.h"
#include "stdio.h"
#include <zephyr/bluetooth/audio/audio.h>
#include <zephyr/kernel.h>
#include <zephyr/zbus/zbus.h>

#include "nrf5340_audio_common.h"
#include "nrf5340_audio_dk.h"
#include "broadcast_source.h"
#include "led.h"
#include "button_assignments.h"
#include "macros_common.h"
#include "audio_system.h"
#include "bt_mgmt.h"

/********************CH101 sensor header files*****************/

#include <zephyr/types.h>
#include <zephyr/kernel.h>
#include <zephyr/drivers/uart.h>
#include <zephyr/usb/usb_device.h>
#include <zephyr/sys/util_macro.h>
#include <zephyr/sys/__assert.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/sys/printk.h>
#include <zephyr/logging/log.h>

#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <soc.h>

#include <zephyr/bluetooth/bluetooth.h>
#include <zephyr/bluetooth/uuid.h>
#include <zephyr/bluetooth/gatt.h>
#include <zephyr/bluetooth/hci.h>

#include <bluetooth/services/nus.h>
#include <dk_buttons_and_leds.h>
#include <zephyr/settings/settings.h>
#include <stdio.h>

#include "ultrasonic_sensor.h"
#include "chirp_board_config.h"
#include "nrfboard/chbsp_chirp.h"

//_________________________-Sensor header files --------

#include <zephyr/drivers/i2c.h>
#include <zephyr/drivers/sensor.h>
#include <nrfx_twim.h>
#include <zephyr/audio/dmic.h>
// #include <src/invn/soniclib/soniclib.h>
// #include <src/nrfboard/chbsp_chirp.h>

// #define LOG_MODULE_NAME main_init
// LOG_MODULE_REGISTER(main, LOG_MODULE_NAME);s


LOG_MODULE_REGISTER(main, CONFIG_MAIN_LOG_LEVEL);
ZBUS_SUBSCRIBER_DEFINE(button_evt_sub, CONFIG_BUTTON_MSG_SUB_QUEUE_SIZE);

ZBUS_MSG_SUBSCRIBER_DEFINE(le_audio_evt_sub);

ZBUS_CHAN_DECLARE(button_chan);
ZBUS_CHAN_DECLARE(le_audio_chan);
ZBUS_CHAN_DECLARE(bt_mgmt_chan);
ZBUS_CHAN_DECLARE(sdu_ref_chan);

ZBUS_OBS_DECLARE(sdu_ref_msg_listen);

static struct k_thread button_msg_sub_thread_data;
static struct k_thread le_audio_msg_sub_thread_data;

static k_tid_t button_msg_sub_thread_id;
static k_tid_t le_audio_msg_sub_thread_id;

K_THREAD_STACK_DEFINE(button_msg_sub_thread_stack, CONFIG_BUTTON_MSG_SUB_STACK_SIZE);
K_THREAD_STACK_DEFINE(le_audio_msg_sub_thread_stack, CONFIG_LE_AUDIO_MSG_SUB_STACK_SIZE);

static enum stream_state strm_state = STATE_PAUSED;

/*********************GPIO Configuration***************/
 #define RESET_PIN 14

 #define GPIO_PORT DT_LABEL(DT_NODELABEL(gpio1))

/*************CH101 variable init start**********************/

// #define I2C0_NODE DT_NODELABEL(chultra)
// const struct device *const ch101_sensor = DEVICE_DT_GET(I2C0_NODE);
// static const struct i2c_dt_spec dev_i2c0 = I2C_DT_SPEC_GET(I2C0_NODE);


#define CONFIG_CH101_ENABLED    1
#define CONFIG_IIS_SCANNER	0
#define CH101GPIO_NODE DT_PATH(ch101gpio)

#define CH101_INT   DT_ALIAS(ch101_int)
#if !DT_NODE_HAS_STATUS(CH101_INT, okay)
#error "Unsupported board: ch101-int devicetree alias is not defined"
#endif

static const struct gpio_dt_spec ch101_int = GPIO_DT_SPEC_GET_OR(CH101_INT, gpios, {0});
static struct gpio_callback ch101_int_cb_data;

#define GPIO_SPEC_AND_COMMA(CH101GPIO_NODE) GPIO_DT_SPEC_GET(CH101GPIO_NODE, gpios),

const struct gpio_dt_spec ch101gpio[] = {
#if DT_NODE_EXISTS(CH101GPIO_NODE)
    DT_FOREACH_CHILD(CH101GPIO_NODE, GPIO_SPEC_AND_COMMA)
#endif
};

static void ch101_int_callback(const struct device *dev, struct gpio_callback *cb, uint32_t pins);
static void ch101_data_ready_work_handler(struct k_work *work);
static K_WORK_DEFINE(ch101_data_ready_work, ch101_data_ready_work_handler);


/*************CH101 variable init end**********************/



/* Function for handling all stream state changes */
static void stream_state_set(enum stream_state stream_state_new)
{
	strm_state = stream_state_new;
}

/**
 * @brief	Handle button activity.
 */
static void button_msg_sub_thread(void)
{
	int ret;
	const struct zbus_channel *chan;

	while (1) {
		ret = zbus_sub_wait(&button_evt_sub, &chan, K_FOREVER);
		ERR_CHK(ret);

		struct button_msg msg;

		ret = zbus_chan_read(chan, &msg, ZBUS_READ_TIMEOUT_MS);
		ERR_CHK(ret);
		LOG_DBG("Got btn evt from queue - id = %d, action = %d", msg.button_pin,
			msg.button_action);

		if (msg.button_action != BUTTON_PRESS) {
			LOG_WRN("Unhandled button action");
			return;
		}

		switch (msg.button_pin) {
		case BUTTON_PLAY_PAUSE:
			if (strm_state == STATE_STREAMING) {
				ret = broadcast_source_stop();
				if (ret) {
					LOG_WRN("Failed to stop broadcaster: %d", ret);
				}
			} else if (strm_state == STATE_PAUSED) {
				ret = broadcast_source_start(NULL);
				if (ret) {
					LOG_WRN("Failed to start broadcaster: %d", ret);
				}
			} else {
				LOG_WRN("In invalid state: %d", strm_state);
			}

			break;

		case BUTTON_4:
			if (IS_ENABLED(CONFIG_AUDIO_TEST_TONE)) {
				if (strm_state != STATE_STREAMING) {
					LOG_WRN("Not in streaming state");
					break;
				}

				ret = audio_system_encode_test_tone_step();
				if (ret) {
					LOG_WRN("Failed to play test tone, ret: %d", ret);
				}

				break;
			}

			break;

		default:
			LOG_WRN("Unexpected/unhandled button id: %d", msg.button_pin);
		}

		STACK_USAGE_PRINT("button_msg_thread", &button_msg_sub_thread_data);
	}
}

/**
 * @brief	Handle Bluetooth LE audio events.
 */
static void le_audio_msg_sub_thread(void)
{
	int ret;
	const struct zbus_channel *chan;

	while (1) {
		struct le_audio_msg msg;

		ret = zbus_sub_wait_msg(&le_audio_evt_sub, &chan, &msg, K_FOREVER);
		ERR_CHK(ret);

		LOG_DBG("Received event = %d, current state = %d", msg.event, strm_state);

		switch (msg.event) {
		case LE_AUDIO_EVT_STREAMING:
			LOG_DBG("LE audio evt streaming");

			audio_system_encoder_start();

			if (strm_state == STATE_STREAMING) {
				LOG_DBG("Got streaming event in streaming state");
				break;
			}

			audio_system_start();
			stream_state_set(STATE_STREAMING);
			ret = led_blink(LED_APP_1_BLUE);
			ERR_CHK(ret);

			break;

		case LE_AUDIO_EVT_NOT_STREAMING:
			LOG_DBG("LE audio evt not_streaming");

			audio_system_encoder_stop();

			if (strm_state == STATE_PAUSED) {
				LOG_DBG("Got not_streaming event in paused state");
				break;
			}

			stream_state_set(STATE_PAUSED);
			audio_system_stop();
			ret = led_on(LED_APP_1_BLUE);
			ERR_CHK(ret);

			break;

		default:
			LOG_WRN("Unexpected/unhandled le_audio event: %d", msg.event);

			break;
		}

		STACK_USAGE_PRINT("le_audio_msg_thread", &le_audio_msg_sub_thread_data);
	}
}

/**
 * @brief	Create zbus subscriber threads.
 *
 * @return	0 for success, error otherwise.
 */
static int zbus_subscribers_create(void)
{
	int ret;

	button_msg_sub_thread_id = k_thread_create(
		&button_msg_sub_thread_data, button_msg_sub_thread_stack,
		CONFIG_BUTTON_MSG_SUB_STACK_SIZE, (k_thread_entry_t)button_msg_sub_thread, NULL,
		NULL, NULL, K_PRIO_PREEMPT(CONFIG_BUTTON_MSG_SUB_THREAD_PRIO), 0, K_NO_WAIT);
	ret = k_thread_name_set(button_msg_sub_thread_id, "BUTTON_MSG_SUB");
	if (ret) {
		LOG_ERR("Failed to create button_msg thread");
		return ret;
	}

	le_audio_msg_sub_thread_id = k_thread_create(
		&le_audio_msg_sub_thread_data, le_audio_msg_sub_thread_stack,
		CONFIG_LE_AUDIO_MSG_SUB_STACK_SIZE, (k_thread_entry_t)le_audio_msg_sub_thread, NULL,
		NULL, NULL, K_PRIO_PREEMPT(CONFIG_LE_AUDIO_MSG_SUB_THREAD_PRIO), 0, K_NO_WAIT);
	ret = k_thread_name_set(le_audio_msg_sub_thread_id, "LE_AUDIO_MSG_SUB");
	if (ret) {
		LOG_ERR("Failed to create le_audio_msg thread");
		return ret;
	}

	ret = zbus_chan_add_obs(&sdu_ref_chan, &sdu_ref_msg_listen, ZBUS_ADD_OBS_TIMEOUT_MS);
	if (ret) {
		LOG_ERR("Failed to add timestamp listener");
		return ret;
	}

	return 0;
}

/**
 * @brief	Zbus listener to receive events from bt_mgmt.
 *
 * @param[in]	chan	Zbus channel.
 *
 * @note	Will in most cases be called from BT_RX context,
 *		so there should not be too much processing done here.
 */
static void bt_mgmt_evt_handler(const struct zbus_channel *chan)
{
	int ret;
	const struct bt_mgmt_msg *msg;

	msg = zbus_chan_const_msg(chan);

	switch (msg->event) {
	case BT_MGMT_EXT_ADV_WITH_PA_READY:
		LOG_INF("Ext adv ready");

		ret = broadcast_source_start(msg->ext_adv);
		if (ret) {
			LOG_ERR("Failed to start broadcaster: %d", ret);
		}

		break;

	default:
		LOG_WRN("Unexpected/unhandled bt_mgmt event: %d", msg->event);
		break;
	}
}

ZBUS_LISTENER_DEFINE(bt_mgmt_evt_listen, bt_mgmt_evt_handler);

/**
 * @brief	Link zbus producers and observers.
 *
 * @return	0 for success, error otherwise.
 */
static int zbus_link_producers_observers(void)
{
	int ret;

	if (!IS_ENABLED(CONFIG_ZBUS)) {
		return -ENOTSUP;
	}

	ret = zbus_chan_add_obs(&button_chan, &button_evt_sub, ZBUS_ADD_OBS_TIMEOUT_MS);
	if (ret) {
		LOG_ERR("Failed to add button sub");
		return ret;
	}

	ret = zbus_chan_add_obs(&le_audio_chan, &le_audio_evt_sub, ZBUS_ADD_OBS_TIMEOUT_MS);
	if (ret) {
		LOG_ERR("Failed to add le_audio sub");
		return ret;
	}

	ret = zbus_chan_add_obs(&bt_mgmt_chan, &bt_mgmt_evt_listen, ZBUS_ADD_OBS_TIMEOUT_MS);
	if (ret) {
		LOG_ERR("Failed to add bt_mgmt listener");
		return ret;
	}

	return 0;
}

uint8_t stream_state_get(void)
{
	return strm_state;
}

void streamctrl_send(void const *const data, size_t size, uint8_t num_ch)
{
	int ret;
	static int prev_ret;

	struct le_audio_encoded_audio enc_audio = {.data = data, .size = size, .num_ch = num_ch};

	if (strm_state == STATE_STREAMING) {
		ret = broadcast_source_send(enc_audio);

		if (ret != 0 && ret != prev_ret) {
			if (ret == -ECANCELED) {
				LOG_WRN("Sending operation cancelled");
			} else {
				LOG_WRN("Problem with sending LE audio data, ret: %d", ret);
			}
		}

		prev_ret = ret;
	}
}

static int uart_init(void)
{
	int err;
	int pos;
	struct uart_data_t *rx;
	struct uart_data_t *tx;

	if (!device_is_ready(uart)) {
		return -ENODEV;
	}

	if (IS_ENABLED(CONFIG_USB_DEVICE_STACK)) {
		err = usb_enable(NULL);
		if (err && (err != -EALREADY)) {
			LOG_ERR("Failed to enable USB");
			return err;
		}
	}

	rx = k_malloc(sizeof(*rx));
	if (rx) {
		rx->len = 0;
	} else {
		return -ENOMEM;
	}

	k_work_init_delayable(&uart_work, uart_work_handler);


	if (IS_ENABLED(CONFIG_BT_NUS_UART_ASYNC_ADAPTER) && !uart_test_async_api(uart)) {
		/* Implement API adapter */
		uart_async_adapter_init(async_adapter, uart);
		uart = async_adapter;
	}

	err = uart_callback_set(uart, uart_cb, NULL);
	if (err) {
		k_free(rx);
		LOG_ERR("Cannot initialize UART callback");
		return err;
	}

	if (IS_ENABLED(CONFIG_UART_LINE_CTRL)) {
		LOG_INF("Wait for DTR");
		while (true) {
			uint32_t dtr = 0;

			uart_line_ctrl_get(uart, UART_LINE_CTRL_DTR, &dtr);
			if (dtr) {
				break;
			}
			/* Give CPU resources to low priority threads. */
			k_sleep(K_MSEC(100));
		}
		LOG_INF("DTR set");
		err = uart_line_ctrl_set(uart, UART_LINE_CTRL_DCD, 1);
		if (err) {
			LOG_WRN("Failed to set DCD, ret code %d", err);
		}
		err = uart_line_ctrl_set(uart, UART_LINE_CTRL_DSR, 1);
		if (err) {
			LOG_WRN("Failed to set DSR, ret code %d", err);
		}
	}

	tx = k_malloc(sizeof(*tx));

	if (tx) {
		pos = snprintf(tx->data, sizeof(tx->data),
			       "UART service initialized\r\n");

		if ((pos < 0) || (pos >= sizeof(tx->data))) {
			k_free(rx);
			k_free(tx);
			LOG_ERR("snprintf returned %d", pos);
			return -ENOMEM;
		}

		tx->len = pos;
	} else {
		k_free(rx);
		return -ENOMEM;
	}

	err = uart_tx(uart, tx->data, tx->len, SYS_FOREVER_MS);
	if (err) {
		k_free(rx);
		k_free(tx);
		LOG_ERR("Cannot display welcome message (err: %d)", err);
		return err;
	}

	err = uart_rx_enable(uart, rx->data, sizeof(rx->data), UART_WAIT_FOR_RX);
	if (err) {
		LOG_ERR("Cannot enable uart reception (err: %d)", err);
		/* Free the rx buffer only because the tx buffer will be handled in the callback */
		k_free(rx);
	}

	return err;
}

static void configure_gpio(void)
{
	int err;

    printk("GPIO Initialization\n");

	LOG_DBG("GPIO Initialization\n");

	LOG_INF("GPIO Initialization\n");
    
#ifdef CONFIG_BT_NUS_SECURITY_ENABLED
	err = dk_buttons_init(button_changed);
	if (err) {
		LOG_ERR("Cannot init buttons (err: %d)", err);
	}
#endif /* CONFIG_BT_NUS_SECURITY_ENABLED */

	err = dk_leds_init();
	if (err) {
		LOG_ERR("Cannot init LEDs (err: %d)", err);
	}

       for (size_t i = 0; i < ARRAY_SIZE(ch101gpio); i++) {
         err = gpio_pin_configure_dt(&ch101gpio[i], GPIO_OUTPUT);
        if (err) {
            LOG_ERR("Cannot configure ch101 gpio");         
        }
    }

    if(!err) {
        
        err = gpio_pin_set_dt(&ch101gpio[0], 0);
        if (err) {
            LOG_ERR("Cannot write ch101 prog gpio");            
        }
        err = gpio_pin_set_dt(&ch101gpio[1], 1);
        if (err) {
            LOG_ERR("Cannot write ch101 reset gpio");           
        }
		err = gpio_pin_set_dt(&ch101gpio[2], 1);
        if (err) {
            LOG_ERR("Cannot write ch101 int-dir gpio");           
        }
    }
}

/*******************CH101 Initialization*********************/

static void ch101_int_callback(const struct device *dev, struct gpio_callback *cb, uint32_t pins)
{

	ARG_UNUSED(dev);
	ARG_UNUSED(cb);
	ARG_UNUSED(pins);

	gpio_pin_interrupt_configure_dt(&ch101_int, GPIO_INT_DISABLE);

	/* Submit IMU data ready work */					
	k_work_submit(&ch101_data_ready_work);
}

static void ch101_data_ready_work_handler(struct k_work *work)
{
	ch101_init();
	while(1);
}

int ch_interrupts_init(void)
{
	int ret;
	
	if (!gpio_is_ready_dt(&ch101_int)) {
		LOG_ERR("Cannot get pointer to ch101_int device");
		return -EINVAL;
	}

	ret = gpio_pin_configure_dt(&ch101_int, GPIO_INPUT);
	if (ret < 0) {
		LOG_ERR("Could not configure ch101_int pin");
		return ret;
	}
	
	ret =  gpio_pin_interrupt_configure_dt(&ch101_int, GPIO_INT_EDGE_TO_ACTIVE);
	if (ret < 0) {
		LOG_ERR("Could not configure ch101_int");
		return ret;
	}

	gpio_init_callback(&ch101_int_cb_data, ch101_int_callback,  BIT(ch101_int.pin));
	if (gpio_add_callback(ch101_int.port, &ch101_int_cb_data) < 0) {
		LOG_ERR("Could not set ch101_int gpio callback");
		return -EIO;
	}
	
	if(!ret) LOG_INF("ch101_interrupts_init success");
	return ret;
}

int main(void)
{
	int err = 0;
	int ret;
	int blink_status = 0;

	const struct device *gpio_dev;

	// static const struct bt_data *ext_adv;
	// static const struct bt_data *per_adv;
	gpio_pin_set(gpio_dev);


    // Get the device binding for GPIO
    gpio_dev = device_get_binding(GPIO_PORT);
    if (!gpio_dev) {
        printk("Failed to bind to GPIO device\n");
        return;
    }

    // Configure the reset pin as an output
     ret = gpio_pin_configure(gpio_dev, RESET_PIN, GPIO_OUTPUT_ACTIVE);
    if (ret < 0) {
        printk("Failed to configure reset pin\n");
        return;
    }

    // Set the reset pin high
    gpio_pin_set(gpio_dev, RESET_PIN, 1);  // 1 sets the pin high
    printk("Reset pin set to high\n");


	printk("\n\n");
	// printf("*******	WELCOME TO MINI-CPS	*******\r\n");
	LOG_INF("*******	CH101 Testing	*******\r\n");
	LOG_INF("SW_version: v1.4\tHW_version:EVB\n");

	// printf("*******	CH101 Testing	*******\r\n");
	// printf("SW_version: v1.4\tHW_version:EVB\n");
	
    printk("*******	CH101 Testing	*******\r\n");
	printk("SW_version: v1.4\tHW_version:EVB\n");

    printk("\n\n");
	
	// k_sleep(K_MSEC(500));

    configure_gpio();	
	
	// Laser_sensor_i2c_init();
	// Mic_i2c_init();
	err = uart_init();
	if (err) {
		error();
	}

    #ifdef CONFIG_IIS_SCANNER
	I2C_scanner();
	#endif

    /**************ch101 sensor init***************** */
	#ifdef CONFIG_CH101_ENABLED
	gpio_pin_interrupt_configure_dt(&ch101_int, GPIO_INT_DISABLE);
		if (ch_interrupts_init())
		{
            
		}
		else{
			// LOG_INF("ch101_init\n");
			ch101_init();
		}
    
	#endif
	/**************ch101 sensor end***************** */

	// LOG_DBG("nRF5340 APP core started");

	// ret = nrf5340_audio_dk_init();
	// ERR_CHK(ret);

	// ret = nrf5340_audio_common_init();
	// ERR_CHK(ret);

	// size_t ext_adv_size = 0;
	// size_t per_adv_size = 0;

	// ret = zbus_subscribers_create();
	// ERR_CHK_MSG(ret, "Failed to create zbus subscriber threads");

	// ret = zbus_link_producers_observers();
	// ERR_CHK_MSG(ret, "Failed to link zbus producers and observers");

	// ret = broadcast_source_enable();
	// ERR_CHK_MSG(ret, "Failed to enable broadcaster");

	// ret = audio_system_config_set(
	// 	bt_audio_codec_cfg_freq_to_freq_hz(CONFIG_BT_AUDIO_PREF_SAMPLE_RATE_VALUE),
	// 	CONFIG_BT_AUDIO_BITRATE_BROADCAST_SRC, VALUE_NOT_SET);
	// ERR_CHK_MSG(ret, "Failed to set sample- and bitrate");

	// broadcast_source_adv_get(&ext_adv, &ext_adv_size, &per_adv, &per_adv_size);

	// ret = bt_mgmt_adv_start(ext_adv, ext_adv_size, per_adv, per_adv_size, false);
	// ERR_CHK_MSG(ret, "Failed to start advertiser");

	return 0;
}
