/**
  ******************************************************************************
  * @file    ultrasonic_sensor.c
  * @author  Argus Team
  * @brief   ch101 driver file
  ******************************************************************************
  * @attention
  *
  * <h2><center>&copy; Copyright (c) 2024 STMicroelectronics.
  * All rights reserved.</center></h2>
  *
  * This software component is licensed by ST under BSD 3-Clause license,
  * the "License"; You may not use this file except in compliance with the
  * License. You may obtain a copy of the License at:
  *                        opensource.org/licenses/BSD-3-Clause
  *
  ******************************************************************************
  */
 
 #include <zephyr/kernel.h>
 #include <zephyr/device.h>
 #include <zephyr/drivers/i2c.h>
 #include "ultrasonic_sensor.h"

 #include "invn/soniclib/soniclib.h"
 #include <invn/soniclib/details/ch_common.h>
 #include <invn/soniclib/ch_rangefinder.h>
 #include <invn/soniclib/sensor_fw/ch101/ch101_gpr.h>


 #include <zephyr/logging/log.h>
 #define LOG_MODULE_NAME ch101_sensor
LOG_MODULE_REGISTER(LOG_MODULE_NAME);

ch_dev_t ch101_dev;
ch_group_t argush;

ch_config_t ch101_config;

void print_senosr_info(void)
{
    uint16_t partnumber = ch_get_part_number(&ch101_dev);
    uint8_t dev_num = ch_get_dev_num(&ch101_dev);
    uint8_t i2c_prog_addr = ch_get_i2c_address(&ch101_dev);
    uint32_t frequency = ch_get_frequency(&ch101_dev);
    uint16_t bandwidth_ch101 = ch_get_bandwidth(&ch101_dev);
    uint16_t result = ch_get_rtc_cal_result(&ch101_dev);
    // char vfw = ch_get_fw_version_string(&ch101_dev);

    // uint8_t i2c_app_addr = 
    LOG_INF("Sensor information\n");
    LOG_INF("Sensor partnumber:CH%d\n",partnumber);
    LOG_INF("Sensor dev_num:%d\n",dev_num);
    LOG_INF("Sensor address:0x%x\n",i2c_prog_addr);
    LOG_INF("Sensor frequency(Hz):%d\n",frequency);
    LOG_INF("Sensor bandwidth(Hz):%d\n",bandwidth_ch101);
    LOG_INF("RTC Cal (lsb @ ms):%d\n",result);
    LOG_INF("Sensor fw version: %s\n",ch_get_fw_version_string(&ch101_dev));

}

void set_free_run_mode(uint8_t range, uint16_t timeinterval)
{
    int ret=0;
    // Set the free run mode
    ch101_config.max_range = range;
    ch101_config.sample_interval = timeinterval;
    ch101_config.mode = CH_MODE_FREERUN;

    ret = ch_set_config(&ch101_dev,&ch101_config);
     if (ret == 0) {
        LOG_INF("ch_set_config successful\n");
         }
         else
          LOG_ERR("ch_set_config failed:%d\n",ret);
    

}

int ch101_init()
{
    uint8_t dev_num = 0;
    ch_fw_init_func_t fw_init_func = ch101_gpr_init;   
    // memset((ch_group_t*)argus, 0x00, sizeof(ch_group_t));
    argush.device[0] = &ch101_dev;
    argush.num_ports = 1;
    argush.num_buses = 1;
    argush.rtc_cal_pulse_ms = 100;
    argush.rtc_cal_type = CH_RTC_CAL_BUS;  /* Calibrate RTC clocks using I/O bus clock (no pulse generation) */
   
    // Call ch_init function sonicLib
    int result = ch_init(&ch101_dev, &argush, dev_num, fw_init_func);       
  
    if (result == 0) {             
        //Group start
        LOG_INF("ch_init successful\n");        
        // LOG_INF("ch_group_start with delay of 100mS\n");
        chbsp_delay_ms(100);
        result = ch_group_start(&argush);                
        if (result == 0) {
            LOG_INF("ch_group_start successful\n");
            // print_senosr_info();
            //set free run configuration
            set_free_run_mode(1200,100);
        }
        else
          LOG_ERR("ch_group_start failed:%d\n",result);      
       
    }
    else{
        LOG_ERR("ch_init is failed\n");
    }
    return result;
}

int i2c_check()
{

 if (!device_is_ready(i2c_ch101)) {
		LOG_ERR("i2c_ch101 is not ready!\n");
         		return 1;
	}
    else 
    LOG_INF("i2c_ch101 is ready!\n");      
    return 0;
}

int custom_i2c_write(uint8_t address,uint8_t *data,uint8_t num_bytes)
{
 int ret=0;	
    struct i2c_msg msg = {
				.buf = (uint8_t *)data,
				.len = num_bytes,
				.flags = I2C_MSG_WRITE};    

	ret = i2c_transfer(i2c_ch101, &msg, 1, (uint16_t)address);
	if (ret != 0) {
		LOG_ERR("chbsp_i2c_write failed : %d\n", ret);
	}      	
	return ret;
}


